/* Blinker
   Copyright 2011-2019 Stefan Schuermans <stefan@blinkenarea.org>
   Copyleft GNU public license - http://www.gnu.org/copyleft/gpl.html
   a blinkenarea.org project */

#ifndef BLINKER_TETRIS_H
#define BLINKER_TETRIS_H

#include <string>
#include <vector>

#include <BlinkenLib/BlinkenFrame.h>

#include "Color.h"
#include "ColorFile.h"
#include "File.h"
#include "Format.h"
#include "FormatFile.h"
#include "Game.h"
#include "Mgrs.h"
#include "Module.h"
#include "NameFile.h"
#include "OpConn.h"
#include "OpConnIf.h"
#include "OpReqIf.h"
#include "OutStreamFile.h"
#include "Time.h"
#include "TimeCallee.h"
#include "UIntFile.h"

namespace Blinker {

/// tetris game
class Tetris: public Game, public OpReqIf
{
protected:
  /// coordinates of a pixel part of a stone
  struct Coord {
    int x;
    int y;
  };

  /// descriptor of a certain rotation of a stone
  struct RotStone {
    Coord pixels[4]; ///< coordinates of the 4 pixels
  };

  /// descriptor of a stone
  struct Stone {
    RotStone rot[4]; ///< rotations of the stone
  };

public:
  /**
   * @brief constructor
   * @param[in] name module name
   * @param[in] mgrs managers
   * @param[in] dirBase base directory
   */
  Tetris(const std::string &name, Mgrs &mgrs, const Directory &dirBase);

  /// virtual destructor
  virtual ~Tetris();

private:
  /// copy constructor disabled
  Tetris(const Tetris &that);

  /// assignment operator disabled
  const Tetris & operator=(const Tetris &that);

public:
  /// check for update of configuration (derived game), return true on update
  virtual bool updateConfigGame();

  /**
   * @brief check if accepting new operator connection is possible
   * @param[in] name operator interface name
   * @return if accepting new connection is possible
   */
  virtual bool acceptNewOpConn(const std::string &name);

  /**
   * @brief new operator connection
   * @param[in] name operator interface name
   * @param[in] pConn operator connection object
   *
   * The new connection may not yet be used for sending inside this callback.
   */
  virtual void newOpConn(const std::string &name, OpConn *pConn);

  /**
   * @brief key command received on operator connection
   * @param[in] pConn operator connection object
   * @param[in] key key that was pressed
   */
  virtual void opConnRecvKey(OpConn *pConn, char key);

  /**
   * @brief play command received on operator connection
   * @param[in] pConn operator connection object
   * @param[in] sound name of sound to play
   */
  virtual void opConnRecvPlay(OpConn *pConn, const std::string &sound);

  /**
   * @brief operator connection is closed
   * @param[in] pConn operator connection object
   *
   * The connection may not be used for sending any more in this callback.
   */
  virtual void opConnClose(OpConn *pConn);

protected:
  /// re-initialize game (e.g. due to config change)
  virtual void reinitialize();

  /// redraw current game image, expected to call sendFrame() at end
  virtual void redraw();

  /// process next time step of game
  virtual void timeStep();

  /// set up a new stone
  void newStone();

  /// set time for next time step of game - or unset if not needed
  void planTimeStep();

  /// check if stone fits at position
  bool checkStoneFit(int stone, int rot, int y, int x) const;

  /// draw a stone to image buffer
  void drawStone(int stone, int rot, int y, int x);

  /// wipe a stone from image buffer (i.e. replace it with background color)
  void wipeStone(int stone, int rot, int y, int x);

  /// set shape of stone to color in image buffer
  void colorStone(int stone, int rot, int y, int x, ColorData const &color);

protected:
  /// stone data
  static Stone const c_stones[7];

  static int const c_stoneCnt; ///< number of stones
  static int const c_rotCnt;   ///< number of rotations per stone
  static int const c_pixelCnt; ///< number of pixels per stone

  /// descriptor for delay value
  static ValueDescr const c_delayDescr;

  ColorFile m_fileStoneColor; ///< color file for stone color
  UIntFile  m_fileDelay;      ///< file for initial delay in ms per frame

  NameFile m_fileStartSound; ///< "start game" sound name file

  ColorData    m_stoneColor; ///< stone color
  unsigned int m_delay;      ///< initial delay in ms per frame (ball speed)

  OpConn *m_pConn; ///< operator connection of player (or NULL)

  int   m_stone; ///< index of current stone
  int   m_rot;   ///< rotation of current stone
  int   m_posX;  ///< x position of current stone
  int   m_posY;  ///< y position of current stone

  /// tetris field (y * m_width + x), -1 for free, >= 0 for pixel from stone
  std::vector<int> m_field;
}; // class Tetris

} // namespace Blinker

#endif // #ifndef BLINKER_TETRIS_H

